@@ -100,6 +100,25 @@ def stepFinished(componentEnvironment, status):
     pass
 
 
+class FMICallException(Exception):
+    """ Raised when an FMI call fails """
+
+    def __init__(self, function: str, status: int):
+
+        if status in range(5):
+            label = ['ok', 'warning', 'discard', 'error', 'fatal', 'pending'][status]
+        else:
+            label = 'illegal return code'
+
+        super().__init__(f"{function} failed with status {status} ({label}).")
+
+        self.function = function
+        "The name of the FMI function"
+
+        self.status = status
+        "The status returned by the FMI function"
+
+
 class _FMU(object):
     """ Base class for all FMUs """
 
@@ -169,6 +188,8 @@ class _FMU(object):
                 if isinstance(v, c_void_p):
                     v = v.value
                 a += hex(0 if v is None else v)
+            elif t == c_bool:
+                a += str(v.value)
             elif t == POINTER(c_uint):
                 # value references
                 if v is None:
@@ -305,7 +326,7 @@ class _FMU1(_FMU):
             if restype == fmi1Status:
                 # check the status code
                 if res > fmi1Warning:
-                    raise Exception("fmi%s failed with status %d." % (fname, res))
+                    raise FMICallException(function=fname, status=res)
 
             return res
 
@@ -500,7 +521,7 @@ class FMU1Slave(_FMU1):
         return self.fmi1CancelStep(self.component)
 
     def doStep(self, currentCommunicationPoint, communicationStepSize, newStep=fmi1True):
-        return self.fmi1DoStep(self.component, currentCommunicationPoint, communicationStepSize, newStep)
+        self.fmi1DoStep(self.component, currentCommunicationPoint, communicationStepSize, newStep)
 
     def getStatus(self, kind):
         value = fmi1Status(fmi1OK)
