@@ -3,7 +3,7 @@
 import pathlib
 from ctypes import *
 from . import free, calloc
-from .fmi1 import _FMU, printLogMessage
+from .fmi1 import _FMU, FMICallException, printLogMessage
 
 
 fmi2Component            = c_void_p
@@ -212,7 +212,7 @@ class _FMU2(_FMU):
             if restype == fmi2Status:  # status code
                 # check the status code
                 if res > fmi2Warning:
-                    raise Exception("%s failed with status %d." % (fname, res))
+                    raise FMICallException(function=fname, status=res)
 
             return res
 
@@ -367,7 +367,7 @@ class _FMU2(_FMU):
         self.fmi2SerializeFMUstate(self.component, state, serializedState, size)
         return serializedState.raw
 
-    def deSerializeFMUstate(self, serializedState, state=fmi2FMUstate()):
+    def deSerializeFMUstate(self, serializedState, state=None):
         """ De-serialize an FMU state
 
         Parameters:
@@ -377,7 +377,8 @@ class _FMU2(_FMU):
         Returns:
             the de-serialized FMU state
         """
-
+        if state is None:
+            state = fmi2FMUstate()
         buffer = create_string_buffer(serializedState, size=len(serializedState))
         self.fmi2DeSerializeFMUstate(self.component, buffer, len(buffer), byref(state))
         return state
@@ -582,24 +583,24 @@ class FMU2Slave(_FMU2):
     def getStatus(self, kind):
         value = fmi2Status(fmi2OK)
         self.fmi2GetStatus(self.component, kind, byref(value))
-        return value
+        return value.value
 
     def getRealStatus(self, kind):
         value = fmi2Real(0.0)
         self.fmi2GetRealStatus(self.component, kind, byref(value))
-        return value
+        return value.value
 
     def getIntegerStatus(self, kind):
         value = fmi2Integer(0)
         self.fmi2GetIntegerStatus(self.component, kind, byref(value))
-        return value
+        return value.value
 
     def getBooleanStatus(self, kind):
         value = fmi2Boolean(fmi2False)
         self.fmi2GetBooleanStatus(self.component, kind, byref(value))
-        return value
+        return bool(value.value)
 
     def getStringStatus(self, kind):
         value = fmi2String(b'')
         self.fmi2GetStringStatus(self.component, kind, byref(value))
-        return value
+        return value.value.decode('utf-8')
