@@ -1,12 +1,10 @@
 import os
+from typing import List, IO, Union
+
+import fmpy
 import numpy as np
 
 
-class ValidationError(Exception):
-
-    pass
-
-
 def read_csv(filename, variable_names=[], validate=True, structured=False):
     """ Read a CSV file that conforms to the FMI cross-check rules
 
@@ -54,7 +52,7 @@ def read_csv(filename, variable_names=[], validate=True, structured=False):
 
     # check if the time is monotonically increasing
     if traj.size > 1 and np.any(np.diff(time) < 0):
-        raise ValidationError("Values in first column (time) are not monotonically increasing")
+        raise Exception("Values in first column (time) are not monotonically increasing")
 
     # get the trajectory names (without the time)
     traj_names = traj.dtype.names[1:]
@@ -62,7 +60,7 @@ def read_csv(filename, variable_names=[], validate=True, structured=False):
     # check if the variable names match the trajectory names
     for variable_name in variable_names:
         if variable_name not in traj_names:
-            raise ValidationError("Trajectory of '" + variable_name + "' is missing")
+            raise Exception("Trajectory of '" + variable_name + "' is missing")
 
     return traj
 
@@ -295,57 +293,52 @@ def create_plotly_figure(result, names=None, events=False, time_unit=None):
     elif time_unit == 'years':
         time /= 365 * 24 * 60 * 60
     else:
-        raise Exception('time_unit must be one of "ns", "us", "ms", "s", "min", "h", "days" or "years" but was "%s".' % time_unit)
+        raise Exception(f'time_unit must be one of "ns", "us", "ms", "s", "min", "h", "days" or "years" but was "{time_unit}".')
 
     if names is None:
         # plot at most 20 signals
         names = result.dtype.names[1:20]
 
-    # one signal per plot
-    # plots = [(None, [name]) for name in names]
-    plots = []
+    fig = make_subplots(rows=len(names), cols=1, shared_xaxes=True)
 
-    for name in names:
-        plots.append((units[name] if name in units else None, [name]))
+    for i, name in enumerate(names):
 
-    fig = make_subplots(rows=len(plots), cols=1, shared_xaxes=True)
+        y = result[name]
+        unit = units.get(name)
 
-    for i, (unit, names) in enumerate(plots):
+        if unit in display_units:
+            display_unit = display_units[unit]
+            y = y * display_unit.factor + display_unit.offset
+            unit = display_unit.name
 
-        for name in names:
+        line = dict(color='#636efa', width=1)
 
-            y = result[name]
+        if y.dtype in [np.float32, np.float64]:
+            trace = go.Scatter(x=time, y=y, name=name, line=line)
+        elif y.dtype == bool:
+            trace = go.Scatter(x=time, y=y.astype(int), name=name, line=line, fill='tozeroy', fillcolor='rgba(0,0,255,0.1)', line_shape='hv')
+            fig['layout'][f'yaxis{i + 1}'].update(tickvals=[0, 1], ticktext=['false', 'true'], range=[-0.1, 1.1], fixedrange=True)
+        else:
+            trace = go.Scatter(x=time, y=y, name=name, line=line, line_shape='hv')
 
-            if unit in display_units:
-                display_unit = display_units[unit]
-                y = y * display_unit.factor + display_unit.offset
-                unit = display_unit.name
+        fig.add_trace(trace, row=i + 1, col=1)
 
-            fig.add_trace(
-                go.Scatter(x=time, y=y,
-                           name=name,
-                           line=dict(color='#636efa', width=1),
-                           fill='tozeroy' if y.dtype == bool else None,
-                           fillcolor='rgba(0,0,255,0.1)'),
-                row=i + 1, col=1)
-
-        title = "%s [%s]" % (name, unit) if unit else name
-
-        fig['layout']['yaxis%d' % (i + 1)].update(title=title)
+        fig['layout'][f'yaxis{i + 1}'].update(title=f"{name} [{unit}]" if unit else name)
 
     if events:
         for t_event in time[np.argwhere(np.diff(time) == 0).flatten()]:
             fig.add_vline(x=t_event, line={'color': '#fbe424', 'width': 1})
 
-    fig['layout']['height'] = 160 * len(plots) + 30 * max(0, 5 - len(plots))
+    fig['layout']['height'] = 160 * len(names) + 30 * max(0, 5 - len(names))
     fig['layout']['margin']['t'] = 30
     fig['layout']['margin']['b'] = 0
     fig['layout']['margin']['r'] = 30
     fig['layout']['plot_bgcolor'] = 'rgba(0,0,0,0)'
-    fig['layout']['xaxis%d' % len(plots)].update(title='time [%s]' % time_unit)
+    fig['layout'][f'xaxis{len(names)}'].update(title=f'time [{time_unit}]')
 
-    fig.update_xaxes(showgrid=True, gridwidth=1, ticklen=0, gridcolor='LightGrey', linecolor='black', showline=True, zeroline=True, zerolinewidth=1, zerolinecolor='LightGrey')
-    fig.update_yaxes(showgrid=True, gridwidth=1, ticklen=0, gridcolor='LightGrey', linecolor='black', showline=True, zerolinewidth=1, zerolinecolor='LightGrey')
+    axes_attrs = dict(showgrid=True, gridwidth=1, ticklen=0, gridcolor='LightGrey', linecolor='black', showline=True, zerolinewidth=1, zerolinecolor='LightGrey')
+    fig.update_xaxes(zeroline=True, **axes_attrs)
+    fig.update_yaxes(**axes_attrs)
 
     fig.update_layout(showlegend=False)
 
@@ -393,8 +386,13 @@ def plot_result(result, reference=None, names=None, filename=None, window_title=
     time = result['time']
 
     if names is None:
-        # plot at most 20 signals
-        names = result.dtype.names[1:20]
+
+        names = []
+
+        # plot at most 20 one-dimensional signals
+        for d in result.dtype.descr[1:]:
+            if len(d) < 3 and len(names) < 20:
+                names.append(d[0])
 
     if len(names) > 0:
 
@@ -543,15 +541,18 @@ def sha256_checksum(filename):
     return sha256.hexdigest()
 
 
-def download_file(url, checksum=None):
-    """ Download a file to the current directory """
+def download_file(url: str, checksum: str = None) -> str:
+    """ Download a file to the current directory
+
+        returns the filename of the downloaded file
+    """
 
     filename = os.path.basename(url)
 
     if checksum is not None and os.path.isfile(filename):
         hash = sha256_checksum(filename)
         if hash.startswith(checksum.lower()):
-            return  # file already exists
+            return filename  # file already exists
 
     import requests
 
@@ -580,6 +581,8 @@ def download_file(url, checksum=None):
         if not hash.startswith(checksum):
             raise Exception("%s has the wrong SHA256 checksum. Expected %s but was %s." % (filename, checksum, hash))
 
+    return filename
+
 
 def download_test_file(fmi_version, fmi_type, tool_name, tool_version, model_name, filename):
     """ Download a file from the Test FMUs repository to the current directory """
@@ -596,11 +599,18 @@ def download_test_file(fmi_version, fmi_type, tool_name, tool_version, model_nam
     url = 'https://github.com/modelica/fmi-cross-check/raw/master/fmus'
     url = '/'.join([url, fmi_version, fmi_type, platform, tool_name, tool_version, model_name, filename])
 
-    download_file(url)
+    return download_file(url)
 
 
-def fmu_info(filename, causalities=['input', 'output']):
-    """ Dump the info for an FMU """
+def fmu_info(filename: Union[str, IO], causalities: List[str] = ['input', 'output']) -> str:
+    """ Dump the info for an FMU
+
+    Parameters:
+        filename     filename of the FMU
+        causalities  the causalities of the variables to include
+
+    Returns the info as a multi line string
+    """
 
     from .model_description import read_model_description
     from . import supported_platforms
@@ -614,37 +624,35 @@ def fmu_info(filename, causalities=['input', 'output']):
     if md.coSimulation is not None:
         fmi_types.append('Co-Simulation')
 
-    l = []
+    l = [f"""
+Model Info
 
-    l.append("")
-    l.append("Model Info")
-    l.append("")
-    l.append("  FMI Version       %s" % md.fmiVersion)
-    l.append("  FMI Type          %s" % ', '.join(fmi_types))
-    l.append("  Model Name        %s" % md.modelName)
-    l.append("  Description       %s" % md.description)
-    l.append("  Platforms         %s" % ', '.join(platforms))
-    l.append("  Continuous States %s" % md.numberOfContinuousStates)
-    l.append("  Event Indicators  %s" % md.numberOfEventIndicators)
-    l.append("  Variables         %s" % len(md.modelVariables))
-    l.append("  Generation Tool   %s" % md.generationTool)
-    l.append("  Generation Date   %s" % md.generationDateAndTime)
+  FMI Version        {md.fmiVersion}
+  FMI Type           {', '.join(fmi_types)}
+  Model Name         {md.modelName}
+  Description        {md.description}
+  Platforms          {', '.join(platforms)}
+  Continuous States  {md.numberOfContinuousStates}
+  Event Indicators   {md.numberOfEventIndicators}
+  Variables          {len(md.modelVariables)}
+  Generation Tool    {md.generationTool}
+  Generation Date    {md.generationDateAndTime}
+"""]
 
     if md.defaultExperiment:
 
         ex = md.defaultExperiment
 
-        l.append("")
         l.append('Default Experiment')
         l.append("")
         if ex.startTime:
-            l.append("  Start Time        %g" % ex.startTime)
+            l.append(f"  Start Time         {ex.startTime}")
         if ex.stopTime:
-            l.append("  Stop Time         %g" % ex.stopTime)
+            l.append(f"  Stop Time          {ex.stopTime}")
         if ex.tolerance:
-            l.append("  Tolerance         %g" % ex.tolerance)
+            l.append(f"  Tolerance          {ex.tolerance}")
         if ex.stepSize:
-            l.append("  Step Size         %g" % ex.stepSize)
+            l.append(f"  Step Size          {ex.stepSize}")
 
     inputs = []
     outputs = []
@@ -658,7 +666,7 @@ def fmu_info(filename, causalities=['input', 'output']):
     l.append("")
     l.append("Variables (%s)" % ', '.join(causalities))
     l.append("")
-    l.append('Name                Causality              Start Value  Unit     Description')
+    l.append('  Name               Causality              Start Value  Unit     Description')
     for v in md.modelVariables:
         if v.causality not in causalities:
             continue
@@ -669,7 +677,7 @@ def fmu_info(filename, causalities=['input', 'output']):
 
         args = ['' if s is None else str(s) for s in [v.name, v.causality, start, unit, v.description]]
 
-        l.append('{:19} {:10} {:>23}  {:8} {}'.format(*args))
+        l.append('  {:18} {:10} {:>23}  {:8} {}'.format(*args))
 
     return '\n'.join(l)
 
@@ -716,21 +724,26 @@ def visual_c_versions():
     return sorted(versions)
 
 
-def compile_dll(model_description, sources_dir, compiler=None):
+def compile_dll(model_description, sources_dir, compiler=None, target_platform=None):
     """ Compile the shared library
 
     Parameters:
-        sources_dir:    directory that contains the FMU's source code
-        compiler:       compiler to use (None: use Visual C on Windows, GCC otherwise)
+        sources_dir    directory that contains the FMU's source code
+        compiler       compiler to use (None: use Visual C on Windows, GCC otherwise)
+        platform       platform to compile for the binary for
     """
 
-    from . import platform, sharedLibraryExtension
+    from . import platform, system
+    import subprocess
+
+    if target_platform is None:
+        target_platform = platform
 
     if model_description.fmiVersion == '1.0':
         raise Exception("FMI 1.0 source FMUs are currently not supported")
 
     if compiler is None:
-        if platform.startswith('win'):
+        if target_platform in ['win32', 'win64', 'x86_64-windows']:
             compiler = 'vc'
         else:
             compiler = 'gcc'
@@ -761,14 +774,21 @@ def compile_dll(model_description, sources_dir, compiler=None):
     if len(source_files) == 0:
         raise Exception("No source files specified in the model description.")
 
-    target = build_configuration.modelIdentifier + sharedLibraryExtension
+    print('Compiling platform binary...')
 
-    print('Compiling %s...' % target)
+    sources = ' '.join(source_files)
+    model_identifier = build_configuration.modelIdentifier
 
     if compiler == 'vc':
 
+        target = model_identifier + '.dll'
+
         vc_versions = visual_c_versions()
 
+        toolset = 'x86_amd64' if target_platform in ['win64', 'x86_64-windows'] else 'x86'
+
+        definitions = ' '.join(f' /D{d}' for d in preprocessor_definitions)
+
         if len(vc_versions) == 0:
             raise Exception("No VisualStudio found")
 
@@ -776,37 +796,39 @@ def compile_dll(model_description, sources_dir, compiler=None):
         vc_version = vc_versions[-1]
 
         if vc_version < 150:
-            command = r'call "%%VS%dCOMNTOOLS%%\..\..\VC\vcvarsall.bat"' % vc_version
+            command = rf'call "%%VS{vc_version}COMNTOOLS%%\..\..\VC\vcvarsall.bat" {toolset}'
         else:
             installation_path = visual_studio_installation_path()
-            command = 'call "' + installation_path + r'\VC\Auxiliary\Build\vcvarsall.bat"'
+            command = rf'call "{installation_path}\VC\Auxiliary\Build\vcvarsall.bat" {toolset}'
 
-        if platform == 'win64':
-            command += ' x86_amd64'
-        else:
-            command += ' x86'
-
-        command += ' && cl /LD /I. /I"%s"' % include_dir
-        for definition in preprocessor_definitions:
-            command += ' /D' + definition
-        command += ' /Fe' + build_configuration.modelIdentifier + ' shlwapi.lib ' + ' '.join(source_files)
+        command += f' && cl /LD /I. /I"{include_dir}" {definitions} /Fe{model_identifier} shlwapi.lib {sources}'
 
     elif compiler == 'gcc':
 
-        command = ''
-        if platform.startswith('win'):
-            command += r'set PATH=C:\MinGW\bin;%%PATH%% && '
-        command += 'gcc -c -I. -I%s' % include_dir
-        if platform in ['linux32', 'linux64']:
-            command += ' -fPIC'
-        for definition in preprocessor_definitions:
-            command += ' -D' + definition
-        command += ' ' + ' '.join(source_files)
-        command += ' && gcc'
-        if platform != 'darwin64':
-            command += ' -static-libgcc'
-        command += ' -shared -o%s *.o' % target
+        definitions = ' '.join(f' -D{d}' for d in preprocessor_definitions)
 
+        if target_platform in ['linux64', 'x86_64-linux']:
+
+            cc = 'gcc'
+            target = model_identifier + '.so'
+
+            if system == 'windows':
+                cc = 'wsl ' + cc
+                output = subprocess.check_output(f"wsl wslpath -a '{include_dir}'")
+                include_dir = output.decode('utf-8').strip()
+
+            command = f'{cc} -c -I. -I{include_dir} -fPIC {definitions} {sources}'
+            command += f' && {cc} -static-libgcc -shared -o{target} *.o -lm'
+
+        elif target_platform in ['darwin64', 'x86_64-darwin']:
+
+            target = model_identifier + '.dylib'
+
+            command = f'gcc -c -I. -I{include_dir} {definitions} {sources}'
+            command += f' && gcc -shared -o{target} *.o -lm'
+
+        else:
+            raise Exception("Unsupported target platform for selected compiler: '%s'" % compiler)
     else:
         raise Exception("Unsupported compiler: '%s'" % compiler)
 
@@ -826,7 +848,7 @@ def compile_dll(model_description, sources_dir, compiler=None):
     return str(dll_path)
 
 
-def compile_platform_binary(filename, output_filename=None):
+def compile_platform_binary(filename, output_filename=None, target_platform=None):
     """ Compile the binary of an FMU for the current platform and add it to the FMU
 
     Parameters:
@@ -842,16 +864,21 @@ def compile_platform_binary(filename, output_filename=None):
 
     model_description = read_model_description(filename)
 
-    binary = compile_dll(model_description=model_description, sources_dir=os.path.join(unzipdir, 'sources'))
+    if target_platform is None:
+        target_platform = platform if model_description.fmiVersion in ['1.0', '2.0'] else platform_tuple
+
+    binary = compile_dll(model_description=model_description,
+                         sources_dir=os.path.join(unzipdir, 'sources'),
+                         target_platform=target_platform)
 
     unzipdir2 = extract(filename)
 
-    platform_dir = os.path.join(unzipdir2, 'binaries', platform if model_description.fmiVersion in ['1.0', '2.0'] else platform_tuple)
+    target_platform_dir = os.path.join(unzipdir2, 'binaries', target_platform)
 
-    if not os.path.exists(platform_dir):
-        os.makedirs(platform_dir)
+    if not os.path.exists(target_platform_dir):
+        os.makedirs(target_platform_dir)
 
-    copyfile(src=binary, dst=os.path.join(platform_dir, os.path.basename(binary)))
+    copyfile(src=binary, dst=os.path.join(target_platform_dir, os.path.basename(binary)))
 
     if output_filename is None:
         output_filename = filename  # overwrite the existing archive
@@ -873,58 +900,82 @@ def compile_platform_binary(filename, output_filename=None):
     rmtree(unzipdir2, ignore_errors=True)
 
 
-def add_remoting(filename):
+def add_remoting(filename, host_platform, remote_platform):
+    """
+        win32 on win64 (SM)
+        linux64 on win64 (WSL + TCP)
+        win64 on linux64 (wine + TCP)
+    """
 
     from . import extract, read_model_description, supported_platforms
     from shutil import copyfile, rmtree
     import zipfile
-    import os
+    from os.path import join, isdir, isfile, normpath, relpath
 
     platforms = supported_platforms(filename)
+    current_dir = os.path.dirname(__file__)
 
-    if 'win32' not in platforms:
-        raise Exception("The FMU does not support the platform \"win32\".")
+    methods = {
+        ('win64', 'win32'): 'sm',
+        ('win64', 'linux64'): 'tcp',
+        ('linux64', 'win64'): 'tcp'
+    }
 
-    if 'win64' in platforms:
-        raise Exception("The FMU already supports \"win64\".")
+    if (host_platform, remote_platform) not in methods:
+        raise Exception("Remoting is not supported for the given combination of host and remote platform.")
+
+    if host_platform in platforms:
+        raise Exception(f"The FMU already supports {host_platform}.")
+
+    if remote_platform not in platforms:
+        raise Exception(f"The FMU does not support {remote_platform}.")
+
+    method = methods[(host_platform, remote_platform)]
 
     model_description = read_model_description(filename)
 
-    current_dir = os.path.dirname(__file__)
-    client = os.path.join(current_dir, 'remoting', 'client.dll')
-    server = os.path.join(current_dir, 'remoting', 'server.exe')
-    license = os.path.join(current_dir, 'remoting', 'license.txt')
+    license = join(current_dir, 'remoting', 'license.txt')
 
-    tempdir = extract(filename)
+    if isdir(filename):
+        tempdir = filename
+    else:
+        tempdir = extract(filename)
 
-    if model_description.coSimulation is not None:
+    if model_description.coSimulation:
         model_identifier = model_description.coSimulation.modelIdentifier
     else:
         model_identifier = model_description.modelExchange.modelIdentifier
 
+    sl_ext = {'linux64': '.so', 'win32': '.dll', 'win64': '.dll'}
+    ex_ext = {'linux64': '', 'win32': '.exe', 'win64': '.exe'}
+
     # copy the binaries & license
-    os.mkdir(os.path.join(tempdir, 'binaries', 'win64'))
-    copyfile(client, os.path.join(tempdir, 'binaries', 'win64', model_identifier + '.dll'))
-    copyfile(server, os.path.join(tempdir, 'binaries', 'win64', 'server.exe'))
-    licenses_dir = os.path.join(tempdir, 'documentation', 'licenses')
-    if not os.path.isdir(licenses_dir):
-        os.mkdir(licenses_dir)
-    copyfile(license, os.path.join(tempdir, 'documentation', 'licenses', 'fmpy-remoting-binaries.txt'))
+    os.makedirs(join(tempdir, 'binaries', host_platform), exist_ok=True)
 
-    # create a new archive from the existing files + remoting binaries
-    with zipfile.ZipFile(filename, 'w', zipfile.ZIP_DEFLATED) as zf:
-        base_path = os.path.normpath(tempdir)
-        for dirpath, dirnames, filenames in os.walk(tempdir):
-            for name in sorted(dirnames):
-                path = os.path.normpath(os.path.join(dirpath, name))
-                zf.write(path, os.path.relpath(path, base_path))
-            for name in filenames:
-                path = os.path.normpath(os.path.join(dirpath, name))
-                if os.path.isfile(path):
-                    zf.write(path, os.path.relpath(path, base_path))
+    copyfile(src=join(current_dir, 'remoting', host_platform, f'client_{method}{sl_ext[host_platform]}'),
+             dst=join(tempdir, 'binaries', host_platform, model_identifier + sl_ext[host_platform]))
 
-    # clean up
-    rmtree(tempdir, ignore_errors=True)
+    copyfile(src=join(current_dir, 'remoting', remote_platform, f'server_{method}{ex_ext[remote_platform]}'),
+             dst=join(tempdir, 'binaries', remote_platform, f'server_{method}{ex_ext[remote_platform]}'))
+
+    licenses_dir = join(tempdir, 'documentation', 'licenses')
+    os.makedirs(licenses_dir, exist_ok=True)
+    copyfile(license, join(tempdir, 'documentation', 'licenses', 'fmpy-remoting-binaries.txt'))
+
+    if not isdir(filename):
+        # create a new archive from the existing files + remoting binaries
+        with zipfile.ZipFile(filename, 'w', zipfile.ZIP_DEFLATED) as zf:
+            base_path = normpath(tempdir)
+            for dirpath, dirnames, filenames in os.walk(tempdir):
+                for name in sorted(dirnames):
+                    path = normpath(join(dirpath, name))
+                    zf.write(path, relpath(path, base_path))
+                for name in filenames:
+                    path = normpath(join(dirpath, name))
+                    if isfile(path):
+                        zf.write(path, relpath(path, base_path))
+        # clean up
+        rmtree(tempdir, ignore_errors=True)
 
 
 def auto_interval(t):
@@ -1130,13 +1181,6 @@ def create_cmake_project(filename, project_dir):
         outfile.write(txt)
 
 
-def _is_string(s):
-    """ Python 2 and 3 compatible type check for strings """
-    
-    import sys
-    return isinstance(s, basestring if sys.version_info[0] == 2 else str)
-
-
 def create_jupyter_notebook(filename, notebook_filename=None):
 
     import nbformat as nbf
@@ -1234,3 +1278,68 @@ def create_jupyter_notebook(filename, notebook_filename=None):
 
     with open(notebook_filename, 'w') as f:
         nbf.write(nb, f)
+
+
+def has_wsl():
+    """ Check if the Windows Subsystem for Linux (WSL) is available """
+
+    if fmpy.system != 'windows':
+        return False
+
+    import subprocess
+
+    try:
+        subprocess.run(['wsl', '--help'])
+        return True
+    except:
+        return False
+
+
+def has_wine64():
+    """ Check if the Wine 64-bit is available """
+
+    if fmpy.system != 'linux':
+        return False
+
+    import subprocess
+
+    try:
+        subprocess.run(['wine64', '--help'])
+        return True
+    except:
+        return False
+
+
+def can_simulate(platforms, remote_platform='auto'):
+
+    from . import platform
+
+    if remote_platform is None:  # remoting disabled
+
+        return platform in platforms, None
+
+    elif remote_platform == 'auto':  # auto remoting
+
+        if platform in platforms:
+            return True, None
+        elif platform == 'win64' and 'win32' in platforms:
+            return True, 'win32'
+        elif has_wsl() and 'linux64' in platforms:
+            return True, 'linux64'
+        elif has_wine64() and 'win64' in platforms:
+            return True, 'win64'
+        else:
+            return False, None
+
+    else:  # specific remoting
+
+        if remote_platform == 'win32' and platform == 'win64':
+            return True, remote_platform
+        elif remote_platform == 'win64' and has_wine64():
+            return True, remote_platform
+        elif remote_platform == 'linux64' and has_wsl():
+            return True, remote_platform
+
+    return False, None
+
+
